@@ -1,5 +1,6 @@
 import time
 from datetime import datetime
+import datetime as datetime_mod
 import os
 import sys
 import dateutil.parser
@@ -10,12 +11,63 @@ except:
     import simplejson as json
 
 
+class IronTokenProvider:
+    def __init__(self, token):
+        self.token = token
+
+    def getToken(self):
+        return self.token
+
+
+class KeystoneTokenProvider:
+    def __init__(self, keystone):
+        self.server = keystone["server"] + ("" if keystone["server"].endswith("/") else "/")
+        self.tenant = keystone["tenant"]
+        self.username = keystone["username"]
+        self.password = keystone["password"]
+        self.token = None
+        self.local_expires_at = None
+
+
+    def getToken(self):
+        if self.local_expires_at:
+            date_diff = time.mktime(datetime.now().timetuple()) - time.mktime(self.local_expires_at.timetuple())
+        else:
+            date_diff = time.mktime(datetime.now().timetuple())
+
+        if self.token is None or date_diff > -10:
+            payload = {
+                'auth': {
+                    'tenantName': self.tenant,
+                    'passwordCredentials': {
+                        'username': self.username,
+                        'password': self.password
+                    }
+                }
+            }
+
+            headers = {'content-type': 'application/json', 'Accept': 'application/json'}
+
+            response = requests.post(self.server + 'tokens', data=json.dumps(payload), headers=headers)
+            result = response.json()
+            token_data = result['access']['token']
+
+            issued_at = dateutil.parser.parse(token_data['issued_at']).replace(tzinfo=None)
+            expires = dateutil.parser.parse(token_data['expires']).replace(tzinfo=None)
+            duration = expires - issued_at
+
+            self.local_expires_at = datetime.now() + duration
+            self.token = token_data['id']
+
+        return self.token
+
+
 class IronClient:
-    __version__ = "1.1.1"
+    __version__ = "1.1.2"
 
     def __init__(self, name, version, product, host=None, project_id=None,
-            token=None, protocol=None, port=None, api_version=None,
-            config_file=None):
+                 token=None, protocol=None, port=None, api_version=None,
+                 config_file=None, keystone=None):
         """Prepare a Client that can make HTTP calls and return it.
 
         Keyword arguments:
@@ -43,6 +95,7 @@ class IronClient:
                 "api_version": None,
                 "project_id": None,
                 "token": None,
+                "keystone": None,
         }
         products = {
                 "iron_worker": {
@@ -70,14 +123,29 @@ class IronClient:
         config = configFromFile(config, config_file, product)
         config = configFromArgs(config, host=host, project_id=project_id,
                 token=token, protocol=protocol, port=port,
-                api_version=api_version)
+                api_version=api_version, keystone=keystone)
 
-        required_fields = ["project_id", "token"]
+        required_fields = ["project_id"]
 
         for field in required_fields:
             if config[field] is None:
-                raise ValueError("No %s set. %s is a required field." % (field,
-                    field))
+                raise ValueError("No %s set. %s is a required field." % (field, field))
+
+        keystone_configured = False
+        if config["keystone"] is not None:
+            keystone_required_keys = ["server", "tenant", "username", "password"]
+            if len(intersect(keystone_required_keys, config["keystone"].keys())) == len(keystone_required_keys):
+                self.token_provider = KeystoneTokenProvider(config["keystone"])
+                keystone_configured = True
+            else:
+                raise ValueError("Missing keystone keys.")
+        elif config["token"] is not None:
+            self.token_provider = IronTokenProvider(config["token"])
+
+        if config["token"] is None and not keystone_configured:
+            raise ValueError("At least one of token or keystone should be specified.")
+
+
 
         self.name = name
         self.version = version
@@ -85,6 +153,7 @@ class IronClient:
         self.host = config["host"]
         self.project_id = config["project_id"]
         self.token = config["token"]
+        self.keystone = config["keystone"]
         self.protocol = config["protocol"]
         self.port = config["port"]
         self.api_version = config["api_version"]
@@ -93,9 +162,7 @@ class IronClient:
                 "Accept": "application/json",
                 "User-Agent": "%s (version: %s)" % (self.name, self.version)
         }
-        if self.token:
-            self.headers["Authorization"] = "OAuth %s" % self.token
-        if self.protocol =="https" and self.port == 443:
+        if self.protocol == "https" and self.port == 443:
             self.base_url = "%s://%s/%s/" % (self.protocol, self.host, self.api_version)
         else:
             self.base_url = "%s://%s:%s/%s/" % (self.protocol, self.host,
@@ -107,6 +174,8 @@ class IronClient:
                     % (self.port, 443))
 
     def _doRequest(self, url, method, body="", headers={}):
+        if self.token or self.keystone:
+            headers["Authorization"] = "OAuth %s" % self.token_provider.getToken()
         if method == "GET":
             r = self.conn.get(url, headers=headers)
         elif method == "POST":
@@ -115,6 +184,8 @@ class IronClient:
             r = self.conn.put(url, data=body, headers=headers)
         elif method == "DELETE":
             r = self.conn.delete(url, data=body, headers=headers)
+        elif method == "PATCH":
+            r = self.conn.patch(url, data=body, headers=headers)
         else:
             raise ValueError("Invalid HTTP method")
         return r
@@ -249,6 +320,22 @@ class IronClient:
         return self.request(url=url, method="PUT", body=body, headers=headers,
                 retry=retry)
 
+    def patch(self, url, body="", headers={}, retry=True):
+        """Execute an HTTP PATCH request and return a dict containing the
+        response and the response status code.
+
+        Keyword arguments:
+        url -- The path to execute the result against, not including the API
+               version or project ID, with no leading /. Required.
+        body -- A string or file object to send as the body of the request.
+                Defaults to an empty string.
+        headers -- HTTP Headers to send with the request. Can overwrite the
+                defaults. Defaults to {}.
+        retry -- Whether exponential backoff should be employed. Defaults
+                 to True.
+        """
+        return self.request(url=url, method="PATCH", body=body, headers=headers,
+                retry=retry)
 
     @staticmethod
     def fromRfc3339(timestamp=None):
@@ -308,3 +395,6 @@ def configFromArgs(config, **kwargs):
         if kwargs[k] is not None:
             config[k] = kwargs[k]
     return config
+
+def intersect(a, b):
+    return list(set(a) & set(b))
\ No newline at end of file
